/* Copyright (c) 2003 The Regents of the University of California.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. All advertising materials mentioning features or use of this software
   must display the following acknowledgement:
     This product includes software developed by the University of
     California, Los Angeles and its contributors.
4. Neither the name of the University nor the names of its contributors
   may be used to endorse or promote products derived from this software
   without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
*/

/* System includes */
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdio.h>
#include <sys/timex.h>
#include <sys/timeb.h>
#include <time.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>
#include <linux/types.h>

/* Our includes */
#include "flow_stats.h"
#include "rl_stats.h"
#include "conn_stats.h"
#include "hash.h"
#include "dward.h"
#include "dward.config"
#include "constants.h"
#include "classifier.h"

int cadd=0, cdel=0;

/* DNS servers */
extern __u32 DNS_servers[10];
extern int num_dns;

/* Flow and connection IDs for red-teaming */
extern int FLOW_ID;
extern int CONN_ID;

/* Flag for absolute time logging */
extern int ABSOLUTE_TIME;

/* Variable for statistics gathering */
extern float Astats;

/* Just a dirty trick to learn attack flow from multicast */
int pending_flow=0;

/* Gather data on allocated memory, useful for checking if 
   everything is freed or there is memory leak */
int classmem = 0;

/* RED TEAM flag */
extern int RED_TEAM;

/* RED TEAM statistics */
extern FILE* dward_coss_stats;
extern FILE* coss_stats;
extern FILE* dward_stats;
extern FILE* dward_stats_parsable;
extern int msgread, msgwrite;

/* File for monitoring traffic from a certain source address */
extern FILE* observed_stats;

/* To count false positives */
extern long false_positives;
extern long false_bad;
extern long classifications;

/* Current time */
extern long timenow;

/* Flag for debugging */
extern int DEBUG;

/* Flag for trace replay */
extern int REPLAY;

/* Flag for increasing rate attack model */
extern int INC;

/* Debug file */
FILE* classdebug;

/* DefCOM flag */ 
extern int DefCOM;

/* Function and flag for multicast communication */
void sendMulticastMessage(char*, int);
extern int COSSACK;

extern struct hash_table *RateLimitHash;

/* Classify flow by looking only at sent/received packet ratio.
   This is future work. */
void threshold_classify(struct hash_entry* he)
{
}

/* Classify based on different parameters */
void multiparam_classify(struct hash_entry* he)
{
  double tcprto, icmprto;
  int rate;
  unsigned long sum;
  struct flow_stats* fs = (struct flow_stats*)(he->data);
  int isAttack = NO_ATTACK;
  double ppc, track=-1;

  fs->old_classification = fs->classification;
  /* DefCOM signal overrides our detection */
  if(DefCOM && fs->DefCOM_attack)
    {
      isAttack=TCP_ATTACK;
      fs->DefCOM_attack = FALSE;
    }
  /* COSSACK signal overrides our detection */
  if(COSSACK && fs->COSSACK_new)
    {
      if (fs->COSSACK_attack)
	isAttack = fs->COSSACK_type;
    }
  /* Should we take into account increasing rate attacks */
  if (INC)
    {
      /* Update statistics */
      if (fs->track_TCP == -1 || fs->min_TCP < fs->track_TCP)
	fs->track_TCP = fs->min_TCP;
      if (fs->track_UDP == -1 || fs->min_UDP < fs->track_UDP)
	fs->track_UDP = fs->min_UDP;
      if (fs->track_ICMP == -1 || fs->min_ICMP < fs->track_ICMP)
	fs->track_ICMP = fs->min_ICMP;
      if (fs->track_OTHER == -1 || fs->min_OTHER < fs->track_OTHER)
	fs->track_OTHER = fs->min_OTHER;
      fs->avg_time ++;
      /* Check if buffer has been filled, and calculate
	 average statistics within the window */
      if (fs->avg_time == AVG_TIME)
	{
	  /* Detect the attack. This should check for UDP too
	     but this is future work. */
	  if (fs->track_TCP>0 || fs->track_OTHER>0)
	    {
	      isAttack = INCREASING_ATTACK;
	      track = fs->track_TCP;
	      printf("Increasing rate attack with %lf and %lf\n",fs->track_TCP, fs->track_OTHER);
	      fflush(stdout);
	    }
	  fs->track_TCP = -1;
	  fs->track_UDP = -1;
	  fs->track_ICMP = -1;
	  fs->track_OTHER = -1;
	  fs->avg_time = 0;
	}
    }
  /* Now do the old checking */
  /* Check different parts of protocol for compliance */
  if (fs->TCP_received_from == 0)
    tcprto = fs->TCP_sent_to;
  else 
    tcprto = (double)fs->TCP_sent_to/fs->TCP_received_from;
  if (fs->ICMP_received_from == 0)
    icmprto = fs->ICMP_sent_to;
  else 
    icmprto = (double)fs->ICMP_sent_to/fs->ICMP_received_from;
  /* Update statistics in smoothed manner */
  if (!fs->observe)
    fs->tcprto = fs->tcprto * ALPHA + (1-ALPHA) * tcprto;
  else
    fs->tcprto = fs->tcprto * (ALPHA1+ALPHA2) + (1-ALPHA1-ALPHA2) * tcprto;
  fs->icmprto = fs->icmprto * ALPHA + (1-ALPHA) * icmprto;
  /* Age connections 
  fs->TCP_conns = fs->TCP_conns * ALPHA;
  fs->UDP_conns = fs->UDP_conns * ALPHA;
  fs->ICMP_conns = fs->ICMP_conns * ALPHA;
  fs->OTHER_conns = fs->OTHER_conns * ALPHA;
  */
  /* Detect various attacks */
  if (fs->tcprto > MAX_TCP_RTO)
    {
      isAttack = TCP_ATTACK;
    }
  else if (fs->icmprto > MAX_ICMP_RTO)
    {
      isAttack = ICMP_ATTACK;
    }
  else 
    {
      ppc=(double)fs->UDP_sent_to/fs->UDP_conns;
      /* Detect only spoofing UDP attack */
      sum = fs->TCP_sent_bytes + fs->UDP_sent_bytes 
	+ fs->ICMP_sent_bytes + fs->OTHER_sent_bytes;
      if (sum && (fs->UDP_conns > MAX_UDP_CONNS 
		  && (double)fs->UDP_sent_to/fs->UDP_conns 
		  < MIN_PACKS_PER_UDP_CONN))
	{
	  __u32 address =  *(__u32*)(he->key);
	  int i;
	  for(i=0;i<num_dns;i++)
	  { 
	    if (address == DNS_servers[i])
	      {
		printf("%x matches %x\n",address,DNS_servers[i]);
		break;
	      }
	    else
		printf("%x does not match %x\n",address,DNS_servers[i]);
	  }
	  if (i==num_dns)
	    {
	      isAttack = UDP_ATTACK;
	    }
	}
      else if (sum && (double)fs->OTHER_sent_bytes/sum > 2*fs->other)
	{
	  isAttack = OTHER_ATTACK;
	}
    }
  Astats = isAttack;
  /* If the attack is detected and there is communication, 
     inform the multicast group consisting of other D-WARD systems
  if (COSSACK && isAttack)
    {
      struct in_addr in;  
      in.s_addr = *(unsigned int*)(he->key);
      sendMulticastMessage(inet_ntoa(in),isAttack);
    }
  */
  /* If flow is attack */
  if (isAttack || (pending_flow == (*(unsigned int*)(he->key))))
    {
      /* Count false positives for replayed traffic */
      false_positives++;
      /* Change flow classification */
      fs->classification = ATTACK;
      fs->compliant = 0;
      /* If the flow is already rate-limited clear the notification */
      if (pending_flow == (*(unsigned int*)(he->key))
	  && find_rate_limit_num(pending_flow) > 0)
	pending_flow = 0;
    }
  /* If flow is compliant */
  else if (!isAttack)
    {
      rate = find_rate_limit_num(*((unsigned int*)he->key));
      /* Flow was not rate-limited */
      if (rate == -2)
	fs->classification = NORMAL;
      else 
	{
	  /* Flow was rate-limited and this might affect its 
	     current classification */
	  int droppedB = getflow(*((unsigned int*)he->key),DROP_BYTES);
	  if (fs->compliant >= COMPLIANT_PERIOD 
	      && droppedB == 0 && fs->classification != ATTACK)
	    {
	      fs->classification = NORMAL;
	    }
	  else if (!droppedB)
	    {
	      fs->classification = SUSPICIOUS;
	      fs->compliant ++;
	    }
	}
      if (rate >= MAX_RATE && fs->classification == NORMAL)
	delete_rate_limit_num(*((unsigned int*)he->key));
    }
  if (DEBUG)
    {
      struct in_addr in;  
      in.s_addr = *(unsigned int*)(he->key);
      if (ABSOLUTE_TIME)
	fprintf(classdebug,"%ld: \tFlow to %s sent %d "
		"rec %d B %d and %d rl %d classified as ",
		time(0),inet_ntoa(in),fs->sent_to,fs->received_from,
		fs->sent_bytes, fs->received_bytes,
		find_rate_limit_num(*((unsigned int*)he->key)));
      else
	fprintf(classdebug,"%ld: \tFlow to %s sent %d "
		"rec %d B %d and %d rl %d classified as ",
		timenow,inet_ntoa(in),fs->sent_to,fs->received_from,
		fs->sent_bytes, fs->received_bytes,
		find_rate_limit_num(*((unsigned int*)he->key)));
      switch(fs->classification)
	{
	case NORMAL:
	  fprintf(classdebug,"NORMAL\n");
	  break;
	case SUSPICIOUS:
	  fprintf(classdebug,"SUSPICIOUS\n");
	  break;
	case ATTACK:
	  fprintf(classdebug,"ATTACK %d\n",isAttack);
	  break;
	}
      fflush(classdebug);
    }  
  if (RED_TEAM)
    {
      /* Just log change in classification */
      if (fs->old_classification != fs->classification)
	{
	  struct in_addr in;  
	  in.s_addr = *(unsigned int*)(he->key);
	  
	  if (fs->ID == 0)
	    {
	      fs->ID = FLOW_ID++;
	    }

	  if (ABSOLUTE_TIME)
	    {
	      fprintf(dward_stats,"%ld: \tFlow to %s ID %d sent %d rec %d TCPrto %lf ICMPrto %lf UDPconns %d UDPppc %lf TCPtrack %lf classified as ",
		      time(0),inet_ntoa(in),fs->ID,fs->sent_to, fs->received_from, fs->tcprto, fs->icmprto, fs->UDP_conns, ppc, track);
	      fprintf(dward_stats_parsable,"C \tF \t%ld \t%s \t%d \t",
		      time(0),inet_ntoa(in),fs->ID);
	    }
	  else
	    {
	      char times[100];
	      gettime(time(0), times);
	      fprintf(dward_stats,"%s: \tFlow to %s ID %d sent %d rec %d TCPrto %lf ICMPrto %lf UDPconns %d UDPppc %lf TCPtrack %lf classified as ",
		      times,inet_ntoa(in),fs->ID,fs->sent_to, fs->received_from, fs->tcprto, fs->icmprto, fs->UDP_conns, ppc, track);
	      fprintf(dward_stats_parsable,"C \tF \t%s \t%s \t%d \t",
		      times,inet_ntoa(in),fs->ID);
	    }
	  switch(fs->classification)
	    {
	    case NORMAL:
	      fprintf(dward_stats,
		      "NORMAL\n");
	      fprintf(dward_stats_parsable,
		      "NORMAL\n");
	      break;
	    case SUSPICIOUS:
	      fprintf(dward_stats,
		      "SUSPICIOUS\n");
	      fprintf(dward_stats_parsable,
		      "SUSPICIOUS\n");
	      break;
	    case ATTACK:
	      switch (isAttack)
		{
		case TCP_ATTACK:
		  fprintf(dward_stats,
			  "TCP ATTACK\n");
		  fprintf(dward_stats_parsable,
			  "TCP ATTACK\n");
		  break;
		case UDP_ATTACK:
		  fprintf(dward_stats,
			  "UDP ATTACK\n");
		  fprintf(dward_stats_parsable,
			  "UDP ATTACK\n");
		  break;
		case ICMP_ATTACK:
		  fprintf(dward_stats,
			  "ICMP ATTACK\n");
		  fprintf(dward_stats_parsable,
			  "ICMP ATTACK\n");
		  break;
		case OTHER_ATTACK:
		  fprintf(dward_stats,
			  "OTHER ATTACK\n");
		  fprintf(dward_stats_parsable,
			  "OTHER ATTACK\n");
		  break;
		case INCREASING_ATTACK:
		  fprintf(dward_stats,
			  "INCREASING RATE ATTACK\n");
		  fprintf(dward_stats_parsable,
			  "INCREASING RATE ATTACK\n");
		  break;
		}
	      break;
	    }
	  fflush(dward_stats);
          fflush(dward_stats_parsable);
	}
    }
  /* This is too much printing
  if (fs->observe && observed_stats)
    {
      struct in_addr in;  
      char times[100];
      gettime(time(0), times);
      in.s_addr = *(unsigned int*)(he->key);
      fprintf(observed_stats, "%s: %s %d %d %d %d %d\n", times, inet_ntoa(in), fs->sent_to, fs->received_from, fs->sent_bytes, fs->received_bytes,fs->classification);
      fflush(observed_stats);
    }
  */
}

/* This is future work */
void hmm_classify(struct hash_entry* he)
{
}

/* Classify connection by looking at specific protocol behavior */
void protocol_classify(struct hash_entry* he)
{
  int droppedp = 0, droppedB = 0;
  double rto;
  struct timeb* tm = (struct timeb*)malloc(sizeof(struct timeb));
  struct conn_stats* cs;
  struct conn_key* key = ((struct conn_key*)he->key);
  double rate;

  classmem += sizeof(struct timeb);
  ftime(tm);
  cs = (struct conn_stats*)(he->data);
  if (cs == 0)
    return;
  cs->old_classification = cs->classification;

  /* Get statistics from kernel only for GOOD connections */
  if (cs->classification == GOOD 
      && cs->sent_to > 0)
    {
      /* Get the drop from the kernel */
      droppedB = getconn(key->localIP,key->foreignIP,
			 key->localport,key->foreignport,DROP_BYTES);
      droppedp = getconn(key->localIP,key->foreignIP,
			 key->localport,key->foreignport,DROP_PACKETS);
    }
  /* printf("Classifying now \n"); */
  fflush(stdout);
  rate = (double)cs->sent_bytes/
    (tm->time + (double)tm->millitm/1000 - cs->timestamp);
  
  /* Calculate current packet ratio */
  if (cs->received_from == 0)
	if (cs->sent_to > 1)
    		rto = cs->sent_to/0.1;
	else
		rto = cs->sent_to;
  else 
    rto= (double)cs->sent_to/cs->received_from;
  
  /* Smooth it */
  cs->rto = cs->rto * ALPHA + (1-ALPHA) * rto;

  /* Check if it is transient connection */
  if ((cs->protocol == TCP && cs->sent_to >= 3) ||
      (cs->protocol == UDP && cs->sent_to >= 1) ||
      (cs->protocol == ICMP && cs->sent_to >= 1))
    cs->new_connection = FALSE;
  if (cs->new_connection == TRUE)
    cs->classification = TRANSIENT;
  else
    {
      /* For ratios we use current and not smoothed value
	 since it gives us faster response */
      cs->new_connection = FALSE;
      if (cs->protocol == TCP)
	{
	  cs->droppedp = droppedp;
	  if (rto>0)
	    {
	      if (rto <= MAX_TCP_RTO)
		{
		  cs->classification = GOOD;
		  
		  cadd++;
		  /* Add it to good connections */
		  addconn(key->localIP,key->foreignIP,
			  key->localport,key->foreignport,cs->protocol);
		  
		}
	      else
		{
		  cs->classification = BAD;
		  cdel++;
		  /* Remove it from good connections */
		  delconn(key->localIP,key->foreignIP,
			  key->localport,key->foreignport);
		  printf("Bad connection rto %f\n",rto);
		  fflush(stdout);
		  if (REPLAY)
		    false_bad++;
		}
	    }
	fflush(stdout);
	}
      else 
	if (cs->protocol == ICMP)
	  {
	    cs->droppedp = droppedp;
	    if (FALSE) //rto <= MAX_ICMP_RTO)
	      {
		cs->classification = GOOD;
		if(cs->classification != cs->old_classification)
		  {
		    cadd++;
		    /* Add it to good connections */
		    addconn(key->localIP,key->foreignIP,
			    key->localport,key->foreignport,cs->protocol);
		  }
	      }
	    else
	      {
		cs->classification = BAD;
		cdel++;
		/* Remove it from good connections */
		delconn(key->localIP,key->foreignIP,
			key->localport,key->foreignport);
		printf("Bad icmp %f\n",rto);
		if (REPLAY)
		  false_bad++;
	      }
	  }
	else if (cs->protocol == UDP)
	  {
	    double rtp_rto;
	    if (cs->received_from == 0)
	      rtp_rto =  cs->sent_to;
	    else
	      rtp_rto = (double)cs->sent_to/cs->received_from;
	    cs->droppedp = droppedp;
	    if (cs->application_state == ERROR ||
		(key->foreignport == RP_PORT && rtp_rto > MAX_RTP_RTO))
	      {
		cs->classification = BAD;
		cdel++;
		/* Remove it from good connections */
		delconn(key->localIP,key->foreignIP,
			key->localport,key->foreignport);
		
		if (REPLAY)
		  false_bad++;
	      }
	    else if (cs->application_state != NOSTATE)
	      {
		cs->classification = GOOD;
		cadd++;
		/* Add it to good connections */
		addconn(key->localIP,key->foreignIP,
			key->localport,key->foreignport,cs->protocol);
	      }
	  }
      free(tm);
      classmem -= sizeof(struct timeb);
    }
  if (DEBUG > 1 && !REPLAY)
    {
      struct in_addr in;  
      char a1[17], a2[17];
      in.s_addr = key->localIP;
      strcpy(a1,inet_ntoa(in));
      in.s_addr = key->foreignIP;
      strcpy(a2,inet_ntoa(in));
      
      if(ABSOLUTE_TIME)
	fprintf(classdebug,
		"%ld: Connection %s:%d %s:%d sent "
		"%d rec %d classified as ",
		time(0),a1,key->localport,a2,key->foreignport, 
		cs->sent_to, cs->received_from);
      else
	fprintf(classdebug,
		"%ld: Connection %s:%d %s:%d sent "
		"%d rec %d classified as ",
		timenow,a1,key->localport,a2,key->foreignport, 
		cs->sent_to, cs->received_from);
      switch(cs->classification)
	{
	case GOOD:
	  fprintf(classdebug,"GOOD\n");
	  break;
	case TRANSIENT:
	  fprintf(classdebug,"TRANSIENT\n");
	  break;
	case BAD:
	  fprintf(classdebug,"BAD\n");
	  break;
	}
      fflush(classdebug);
    }
  if (RED_TEAM)
    {
      if (cs->old_classification != cs->classification)
	{
	  struct in_addr in;  
	  char a1[17], a2[17];
	  in.s_addr = key->localIP;
	  strcpy(a1,inet_ntoa(in));
	  in.s_addr = key->foreignIP;
	  strcpy(a2,inet_ntoa(in));
	  
	  if (cs->ID == 0)
	    {
	      cs->ID = CONN_ID++;
	    }
	  if(ABSOLUTE_TIME)
	    {
	      fprintf(dward_stats,
		      "%ld: \tConnection %s:%d %s:%d ID %d protocol %s sent %d received %d classified as ",
		      time(0),a1,key->localport,a2,key->foreignport, cs->ID,print_proto(cs->protocol),cs->sent_to, cs->received_from);
	      fprintf(dward_stats_parsable,
		      "C \tC \t%ld \t%s \t%d \t%s \t%d \t%d \t%s \t",
		      time(0),a1,key->localport,a2,key->foreignport,cs->ID,print_proto(cs->protocol));
	    }
	  else
	    {
	      char times[100];
	      gettime(time(0), times);
	      fprintf(dward_stats,
		      "%s: \tConnection %s:%d %s:%d ID %d protocol %s sent %d received %d classified as ",
		      times,a1,key->localport,a2,key->foreignport,cs->ID,print_proto(cs->protocol), cs->sent_to, cs->received_from);
	      fprintf(dward_stats_parsable,
		      "C \tC \t%s \t%s \t%d \t%s \t%d \t%d \t%s \t",
		      times,a1,key->localport,a2,key->foreignport,cs->ID,print_proto(cs->protocol));
	    }
	  
	  switch(cs->classification)
	    {
	    case GOOD:
	      fprintf(dward_stats,"GOOD \n");
	      fprintf(dward_stats_parsable, "GOOD\n");
	      break;
	    case TRANSIENT:
     	      fprintf(dward_stats,"TRANSIENT \n");
              fprintf(dward_stats_parsable, "TRANSIENT\n");
              break;
	    case BAD:
	      fprintf(dward_stats,"BAD \n");
              fprintf(dward_stats_parsable, "BAD\n");
              break;
	    }
	fflush(dward_stats);
	fflush(dward_stats_parsable);
	}
    }
}




