/*
Copyright (c) 2003 The Regents of the University of California.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. All advertising materials mentioning features or use of this software
   must display the following acknowledgement:
     This product includes software developed by the University of
     California, Los Angeles and its contributors.
4. Neither the name of the University nor the names of its contributors
   may be used to endorse or promote products derived from this software
   without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE. */

#include <sys/ioctl.h>
#include <pthread.h>
#include <fcntl.h>
#include <stdio.h>
#include <netinet/in.h>
#include <linux/ip.h>
#include <linux/tcp.h>
#include <linux/udp.h>
#include <linux/icmp.h>
#include <stdio.h>
#include <math.h>
#include <asm/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <time.h>
#include <unistd.h>
#include <sys/time.h>

#include "flow_stats.h"
#include "conn_stats.h"
#include "sniff.h"

int monmemory = 0;
long packets;

extern struct hash_table* ConnHash;

/* Replay flag */
extern int REPLAY;

/* Debug file for monitor statistics */
FILE* mondebug;

/* Debug flag */
extern int DEBUG;

/* Statistics flag */
extern int STATS;

/* Current time */
extern long timenow;

int forwarded = 0;

struct timeval tv, pv;

extern __u32 LOCAL_ADDRESS;
extern int LOCAL_MASK;

extern int sniff_fd;
unsigned char buffer[BUF_LEN];
int plen;
int packet_count;

extern __u32 aggressive_address[MAX_TCP_AGGRESSIVE];
extern int num_aggressive;

void parse_packets(unsigned char* buffer, int plen, int packet_count)
{
  unsigned char* pos = buffer;
  __u32 source, dest, seqnum, acknum;
  int sport=0, dport=0;
  __u16 tot_len;
  __u32 temp = ((__u32)(~0)) << (__u32)(32-LOCAL_MASK);

  while(pos < buffer + plen*packet_count)
    {
      int i;
      struct in_addr sin, din;
     /* Take one packet out and parse it */
      struct iphdr *ip = (struct iphdr*)pos;
      enum pack_type ptype;
      source = ip->saddr;
      dest = ip->daddr;
      tot_len = ntohs(ip->tot_len);
      ptype = DP;
      sin.s_addr = source;
      din.s_addr = dest;
      if (ip->protocol == IPPROTO_IP || ip->protocol == IPPROTO_TCP)
	{
	  struct tcphdr *tcp = (struct tcphdr*)(pos + 4*ip->ihl); 
	  sport = ntohs(tcp->source);
	  dport = ntohs(tcp->dest);
	  seqnum = (__u32)ntohl(tcp->seq);
	  acknum = (__u32)ntohl(tcp->ack);
          /* printf("Packet from %s:%d ", inet_ntoa(sin), sport);
	  	   printf(" to %s:%d \n", inet_ntoa(din), dport);
	  fflush(stdout); */
	  /* Define type of packet here */
	  if (tcp->syn)
	    {
	      if (tcp->ack)
		ptype = SYNACK;
	      else
		ptype = SYN;
	    }
	  else if (tcp->fin)
	    ptype = FIN;
	  else if (tcp->rst)
	    ptype = RST;
	  else if (tcp->ack && tot_len <= 52)
	    ptype = ACK;
	  else
	    ptype = DP;
	}
      else if (ip->protocol == IPPROTO_UDP)
	{
	  struct udphdr *udp = (struct udphdr*)(pos + 4*ip->ihl);
          unsigned char *apphdr = (int)udp + 8;
          sport = ntohs(udp->source);
          dport = ntohs(udp->dest);
          if(sport == NTP_PORT || dport == NTP_PORT)
            {
              ptype = NTP;
            }
          if(sport == RP_PORT || dport == RP_PORT)
            {
              ptype = STREAM;
            }
          if(sport == DNS_PORT || dport == DNS_PORT)
            {
              if(*(apphdr+2)&128)
                {
                  ptype = DNSACK;
                }
              else
                {
                  ptype = DNS;
                }
            }
	}
      if (ip->protocol == IPPROTO_ICMP)
	{
	}
      /* Update hashes */
      if (((ntohl(dest) & temp) 
	   != (ntohl(LOCAL_ADDRESS) & temp))
	  && ((ntohl(source) & temp) 
	      == (ntohl(LOCAL_ADDRESS) & temp)))
        {
	  /* outgoing packet, source is local address and destination
	     is foreign address */
	  int class;
	  class = update_conn_num(source,dest,sport,dport,OUT,
				  tot_len,ip->protocol,ptype);
	  update_flow_num(dest,OUT,tot_len, ip->protocol, ip->tos, class, ptype);
	  if (num_aggressive > 0)
	    {
	      int k;
	      for(k=0;k<num_aggressive;k++)
		if (source == aggressive_address[k])
		  monitor_flow(dest);
	    }
	}
      else if 
	(((ntohl(dest) & temp) 
	  == (ntohl(LOCAL_ADDRESS) & temp))
	 && ((ntohl(source) & temp) 
	     != (ntohl(LOCAL_ADDRESS) & temp)))
        {  
	  /* incoming packet, source is foreing address and destination
	     is local address */
	  int class;
	  class = update_conn_num(dest,source,dport,
				  sport,IN,tot_len,ip->protocol,ptype);
	  update_flow_num(source,IN,tot_len,ip->protocol, ip->tos, class, ptype);
	} 
      /* Move on to the next packet */
      pos += plen;
   }
}

int run_monitor()
{
  int total_packets = 0;
  struct timeval tv, pv;
  int times=0;
  int bt = time(0);

  while(TRUE)
    {
      struct sniff s;
      times++;
      s.buffer = buffer;
      s.plen = &plen;
      s.packet_count = &packet_count;
      ioctl(sniff_fd, 0, &s);
      total_packets += packet_count;
      if (packet_count)
	{
	  parse_packets(buffer, plen, packet_count);
	}
    }
}














