/*
Copyright (c) 2003 The Regents of the University of California.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. All advertising materials mentioning features or use of this software
   must display the following acknowledgement:
     This product includes software developed by the University of
     California, Los Angeles and its contributors.
4. Neither the name of the University nor the names of its contributors
   may be used to endorse or promote products derived from this software
   without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE. */

/* System includes */
#include <stdio.h>
#include <stdlib.h>
#include <sys/timex.h>
#include <sys/timeb.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <time.h>

/* Our includes */
#include "rate_limit.h"
#include "rl_stats.h"
#include "flow_stats.h"
#include "conn_stats.h"
#include "hash.h"
#include "dward.h"
#include "dward.config"

/* DefCOM flag */ 
extern int DefCOM;

/* COSSACK flag */
extern int COSSACK;

/* Flag for absolute time logging */
extern int ABSOLUTE_TIME;

/* Gather data on allocated memory, useful for checking if 
   everything is freed or there is memory leak */
int ratemem = 0;

/* Current time from beginning of D-WARD operation */
extern long int timenow;

/* Debug variables */
extern int DEBUG;
extern FILE* rldebug;

/* Statistics variables */
extern int STATS;
extern FILE* rlstats;

/* RED TEAM flag */
extern int RED_TEAM;

/* RED TEAM statistics */
extern FILE* dward_coss_stats;
extern FILE* coss_stats;
extern FILE* dward_stats;
extern FILE* dward_stats_parsable;

/* Function that defines TCP-like rate limit */
void variable_rate_limit(struct hash_entry* he)
{
  struct flow_stats* fs = (struct flow_stats*)(he->data);
  unsigned int address = *(unsigned int*)(he->key);
  int rl;
  int droppedp = 0, droppedB = 0;
  struct timeb* tm = (struct timeb*)malloc(sizeof(struct timeb));
  ratemem += sizeof(struct timeb);
  rl = find_rate_limit_num(address);
  ftime(tm);

  /* First check if DefCOM has overridden our rate limit */
  if (DefCOM && fs->DefCOM_rate > 0)
    {
      rl=fs->DefCOM_rate;
      /* Adjust rate limit in hash - 
	 this will also propagate to kernel */
      update_rate_limit_num(address,rl);
    }
  /* Then check if COSSACK has overridden our rate limit */
  else if (COSSACK && fs->COSSACK_timestamp)
    {
      int oldrl = getflow(address, RL);
      /* If this is repeated message that the attack 
	 has stopped, ignore it */
      if (fs->COSSACK_attack || fs->limited)
	{
	  if (fs->COSSACK_type == UDP_ATTACK ||
	      fs->COSSACK_type == OTHER_ATTACK)
	    {
	      if(fs->COSSACK_attack)
		rl = fs->COSSACK_rate;
	      else
		rl = MAX_RATE;
	      /* Adjust rate limit in hash - 
		 this will also propagate to kernel */
	      update_rate_limit_num(address,rl);
	    }
	  if (RED_TEAM && (oldrl != rl))
	    {
	      struct in_addr in;  
	      in.s_addr = *(unsigned int*)(he->key);
	      if (fs->COSSACK_attack)
		{
		  if (ABSOLUTE_TIME)
		    {
		      fprintf(dward_stats,
			      "%ld: \tAuthoritatively set the rate limit "
			      "for %s ID %d to %d \n",
			      time(0),inet_ntoa(in), fs->ID, rl);
		      fprintf(dward_stats_parsable,
			      "RL \tCS-SET \t%ld \t%s \t%d \t%d \n",
			      time(0),inet_ntoa(in), fs->ID, rl);
		    }
		  else
		    {
		      char times[100];
		      gettime(time(0), times);
		      fprintf(dward_stats,
			    "%s: \tAuthoritatively set the rate limit "
			    "for %s ID %d to %d\n",
			    times,inet_ntoa(in), fs->ID, rl);
		      fprintf(dward_stats_parsable,
			      "RL \tCS-SET \t%s \t%s \t%d \t%d \n",
			      times,inet_ntoa(in), fs->ID, rl);
		    }
		fflush(dward_stats);
		fflush(dward_stats_parsable);  
		fs->limited = TRUE;
		}
	      else
		{
		  if (ABSOLUTE_TIME)
		    {
		      fprintf(dward_stats,
			    "%ld: \tAuthoritatively set the rate limit "
			    "for %s ID %d to MAX_RATE %d\n\t\t"
			    "and will be removed next\n",
			    time(0),inet_ntoa(in), fs->ID, rl);
		      fprintf(dward_stats_parsable,
			      "RL \tCS-MAX \t%ld \t%s \t%d \t%d \n",
			      time(0),inet_ntoa(in), fs->ID, rl);
		    }
		  else
		    {
		      char times[100];
		      gettime(time(0), times);
		      fprintf(dward_stats,
			      "%s: \tAuthoritatively set the rate limit "
			      "for %s ID %d to MAX_RATE %d\n\t\t"
			      "and will be removed next\n",
			      times,inet_ntoa(in),fs->ID, rl);
		      fprintf(dward_stats_parsable,
			      "RL \tCS-MAX \t%s \t%s \t%d \t%d \n",
			      times,inet_ntoa(in), fs->ID, rl);
		    }
		  fflush(dward_stats);
		  fflush(dward_stats_parsable);
		  fs->limited = FALSE;
		}
	    }
	}
      /* We have responded to message,
	 reset flag */
      fs->COSSACK_timestamp--;
      if (fs->COSSACK_timestamp <= 0)
	fs->COSSACK_new = FALSE;
    }
  else
    {
      /* If rate-limit should exist, obtain the exact rate-limit from kernel */
      if (rl != -2)
	{
	  int temprl=getflow(address, RL);
	  /* check if kernel has the most recent rate limit or not */
	  if (fs->classification != ATTACK || temprl <= rl || rl <=0)
	    rl = temprl;
	}
      
      /* If it is still -1 or 0 wait until next time */
      if (rl == -1 || rl == 0)
	{
	  free(tm);
	  ratemem -= sizeof(struct timeb);
	  return;
	}
      
      /* If there is a rate limit */
      if (rl > 0)
	{
	  double factor = 1;
	  double rate;
	  /* Obtain statistics from kernel */
	  fs->fw_packets = fs->sent_to;
	  fs->fw_bytes = fs->sent_bytes;
	  fs->sent_to = getflow(address,SENT_PACKETS);
	  fs->received_from = getflow(address,REC_PACKETS);
	  fs->sent_bytes = getflow(address,SENT_BYTES);
	  fs->received_bytes = getflow(address,REC_BYTES);
	  droppedp = getflow(address,DROP_PACKETS);
	  droppedB = getflow(address,DROP_BYTES);
	  getflow(address,ATT_PACKETS);
	  getflow(address,ATT_BYTES);
	  /* Calculate current rate */
	  rate = (double)fs->sent_bytes;
	  /* If flow is not limited set it to limited */
	  if (RED_TEAM)
	    {
	      if (!fs->limited)
		{
		  struct in_addr in;  
		  in.s_addr = *(unsigned int*)(he->key);
		  if (ABSOLUTE_TIME)
		    {
		      fprintf(dward_stats,
			      "%ld: \tStarting exponential decrease"
			      " of rate limit \n\t\ton flow to %s ID %d "
			      "rate set to %d\n",
			      time(0),inet_ntoa(in), fs->ID, rl);
		      fprintf(dward_stats_parsable,
			      "RL \tDW-ED \t%ld \t%s \t%d \t%d \n",
			      time(0),inet_ntoa(in), fs->ID, rl);
		    }
		  else
		    {
		      char times[100];
		      gettime(time(0), times);
		      fprintf(dward_stats,
			      "%s: \tStarting exponential decrease"
			      " of rate limit \n\ton flow to %s ID %d "
			      "rate set to %d\n",
			      times,inet_ntoa(in), fs->ID, rl);
		      fprintf(dward_stats_parsable,
			      "RL \tDW-ED \t%s \t%s \t%d \t%d \n",
			      times,inet_ntoa(in), fs->ID, rl);
		    }
		 fflush(dward_stats);
		 fflush(dward_stats_parsable);
		}
	    }
	  fs->limited = TRUE;
	  
	  /* Calculate compliance factor */
	  if (fs->sent_bytes+droppedB)
	    factor = (double)fs->sent_bytes/
	      (fs->sent_bytes + droppedB); 	  
	  /* If flow is attack */
	  if(fs->classification == ATTACK)
	    {
	      /* Exponentially decrease rate limit */
	      if (rate < rl)
		{ 
		  if (rate > MIN_RATE * DEC_SPEED/factor) 
		    rl = rate/DEC_SPEED * factor;		  
		  else
		    rl = MIN_RATE;
		}
	      else
		{
		  if (rl > MIN_RATE * DEC_SPEED/factor)
		    rl = rl/DEC_SPEED * factor;
		  else
		    rl = MIN_RATE;
		}
	    }
	  /* If it is suspicious linearly increase rate limit */
	  else if (fs->classification == SUSPICIOUS)
	    {
	      if (droppedB == 0)
		rl += (RATE_INC * factor);
	      if (RED_TEAM)
		{
		  if (fs->old_classification != fs->classification)
		    {
		      struct in_addr in;  
		      in.s_addr = *(unsigned int*)(he->key);
		      if(ABSOLUTE_TIME)
			{
			  fprintf(dward_stats,
				  "%ld: \tStarting slow recovery,"
				  " linear increase"
				  " of rate limit \n\t\ton flow to %s ID %d"
				  "rate set to %d\n",
				  time(0),inet_ntoa(in), fs->ID, rl);
			  fprintf(dward_stats_parsable,
				  "RL \tDW-LI \t%ld \t%s \t%d \t%d \n",
				  time(0),inet_ntoa(in), fs->ID, rl);
			}
		      else
			{
			  char times[100];
			  gettime(time(0), times);
			  fprintf(dward_stats,
				  "%s: \tStarting slow recovery,"
				  " linear increase"
				  " of rate limit \n\ton flow to %s ID %d"
				  "rate set to %d\n",
				  times,inet_ntoa(in), fs->ID, rl);
			  fprintf(dward_stats_parsable,
				  "RL \tDW-LI \t%s \t%s \t%d \t%d \n",
				  times,inet_ntoa(in), fs->ID, rl);
			}
			fflush(dward_stats);
			fflush(dward_stats_parsable);
		    }
		}
	    }
	  else
	    {
	      /* If it is normal exponentialy increase rate limit */
	      if (rl < MAX_RATE/INC_SPEED)
		rl = rl + rl * factor; 
	      else
		rl = MAX_RATE;
	      if (RED_TEAM)
		{
		  if (fs->old_classification != fs->classification)
		    {
		      struct in_addr in;  
		      in.s_addr = *(unsigned int*)(he->key);
		      if(ABSOLUTE_TIME)
			{
			  fprintf(dward_stats,
				  "%ld: \tStarting fast recovery,"
				  " exponential increase"
				  " of rate limit \n\t\ton flow to %s ID %d"
				  "rate set to %d\n",
				  time(0),inet_ntoa(in), fs->ID, rl);
			  fprintf(dward_stats_parsable,
				  "RL \tDW-EI \t%ld \t%s \t%d \t%d \n",
				  time(0),inet_ntoa(in), fs->ID, rl);
			}
		      else
			{
			  char times[100];
			  gettime(time(0), times);
			  fprintf(dward_stats,
				  "%s: \tStarting fast recovery,"
				  " exponential increase"
				  " of rate limit \n\ton flow to %s ID %d"
				  "rate set to %d\n",
				  times,inet_ntoa(in), fs->ID, rl);
			  fprintf(dward_stats_parsable,
				  "RL \tDW-EI \t%s \t%s \t%d \t%d \n",
				  times,inet_ntoa(in), fs->ID, rl);
			}
			fflush(dward_stats);
			fflush(dward_stats_parsable);		    
		  }
		  else if (rl == MAX_RATE)
		    {
		      struct in_addr in;  
		      in.s_addr = *(unsigned int*)(he->key);
		      if(ABSOLUTE_TIME)
			{
			  fprintf(dward_stats,
				  "%ld: \tRate for"
				  " flow to %s ID %d"
				  " reached MAX_RATE %d\n\t\t"
				  "and will be removed next\n",
				  time(0),inet_ntoa(in), fs->ID, rl);
			  fprintf(dward_stats_parsable,
				  "RL \tDW-MAX \t%ld \t%s \t%d \t%d \n",
				  time(0),inet_ntoa(in), fs->ID, rl);
			}
		      else
			{
			  char times[100];
			  gettime(time(0), times);
			  fprintf(dward_stats,
				  "%s: \tRate for"
				  " flow to %s ID %d"
				  " reached MAX_RATE %d\n\t"
				  "and will be removed next\n",
				  times,inet_ntoa(in), fs->ID, rl);
			  fprintf(dward_stats_parsable,
				  "RL \tDW-MAX \t%s \t%s \t%d \t%d \n",
				  times,inet_ntoa(in), fs->ID, rl);
			}
		   	fflush(dward_stats);
			fflush(dward_stats_parsable); 
		   }
		}
	    }
	  /* Adjust rate limit in hash - this will also propagate to kernel */
	  update_rate_limit_num(address,rl);
    	}
      else 
	{
	  /* This is newly added rate limit, set it to default value 
	     of -1 and let kernel define it */
	  if(fs->classification == ATTACK)
	    {
	      rl = -1;
	      /* Adjust rate limit in hash - 
		 this will also propagate to kernel */
	      update_rate_limit_num(address,rl);
	    }
     	}
    }  
  if (rl <= 0 && fs->limited)
    { 
      if (RED_TEAM)
	{
	  struct in_addr in;  
	  in.s_addr = *(unsigned int*)(he->key);
	  if(ABSOLUTE_TIME)
	    {
	      fprintf(dward_stats,"%ld: \tRemoved rate limit"
		      " on flow to %s ID %d\n",
		      time(0),inet_ntoa(in), fs->ID);
	      fprintf(dward_stats_parsable,
		      "RL \tDW-REM \t%ld \t%s \t%d \n",
		      time(0),inet_ntoa(in), fs->ID);
	    }
	  else
	    {
	      char times[100];
	      gettime(time(0), times);
	      fprintf(dward_stats,"%s: \tRemoved rate limit"
		      " on flow to %s ID %d\n",
		      times,inet_ntoa(in),fs->ID);
	      fprintf(dward_stats_parsable,
		      "RL \tDW-REM \t%s \t%s \t%d \n",
		      times,inet_ntoa(in), fs->ID);
	    }
	  fflush(dward_stats);
	}
      fs->limited = FALSE;
    }
  if (STATS)
    {
      struct in_addr in;
      in.s_addr = address;

      if (rl>0)
	{
	  if(ABSOLUTE_TIME)
	    fprintf(rlstats,"%d %d %s\n",time(0),rl,inet_ntoa(in));
	  else
	    {
	      char times[100];
	      gettime(time(0), times);
	      fprintf(rlstats,"%s %d %s\n",times,rl,inet_ntoa(in));
	    }
	}
      fflush(rlstats);
    }
  else 
    if (DEBUG)
      {
	struct in_addr in;
	in.s_addr = address;
	
	fprintf(rldebug,
		"%ld: Updated rate limit for %s to %d\n",
		time(0),inet_ntoa(in),rl);
	fflush(rldebug);
      }
  free(tm);
  ratemem -= sizeof(struct timeb);
}








