/*Copyright (c) 2003 The Regents of the University of California.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. All advertising materials mentioning features or use of this software
   must display the following acknowledgement:
     This product includes software developed by the University of
     California, Los Angeles and its contributors.
4. Neither the name of the University nor the names of its contributors
   may be used to endorse or promote products derived from this software
   without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE. */

/* System includes */
#include <stdio.h>

/* Our includes */
#include "Animation.h"
#include "Common.h"
#include "NetworkUtil.h"


/* Forward/external declarations: */
int GetJavaHandles(JNIEnv *, jobject);
void PassSimDataToJava(jstring, JNIEnv*);

/* ... and some globals */
static int gPort;
static JNIEnv *gJavaEnv = NULL;
static jclass gJavaClass = 0;
static jobject gJavaObj, gobj;
static jmethodID gJavaCallbackMethodID;

/*
 * Initializes string with victim addresses
 */
JNIEXPORT jint JNICALL Java_Animation_NativeInitialize  (JNIEnv *env, jobject obj, jint serverPort, jobjectArray array)
{
  const char *str = (*env)->GetStringUTFChars(env, array, 0);
  gPort = (int) serverPort;
  InitializeStrings((char*)str);
  ConnectToServer();
  (*env)->ReleaseStringUTFChars(env, array, str);
  return 0;
}

/* Get handles used to communicate with Java application */
JNIEXPORT void JNICALL Java_Animation_NativePass (JNIEnv *env, jobject obj, jstring mes)
{
  int s;
  GetJavaHandles(env, obj);
  s = getSocketID();
  PassSimDataToJava(mes,env);
}

/* END OF Native Java fuctions definitions, time for some pure C... */

int GetJavaHandles(JNIEnv *env, jobject obj)
{
    jclass clsTmp;
    if (gJavaEnv == NULL)
	{
	    gJavaEnv = env;
	    gJavaObj = obj;
	}	

    /* Get necessary handles for the Java Callback functions (so that we can
     * invoke a Java method in this C file.  (V)V is the
     * (argument-types)return-type signature for the Java method.  V = void */
    clsTmp = (*env)->GetObjectClass(env, obj);
    gJavaClass = (*env)->NewGlobalRef(env, clsTmp);
    gJavaCallbackMethodID=(*env)->GetMethodID(env, clsTmp, "UpdateColors", "(Ljava/lang/String;)V");
    if (gJavaCallbackMethodID==0)
	fprintf(stderr,"Can not locate appropriate method");
    obj=(*env)->AllocObject(env,gJavaClass);
    gobj=obj;
    return 1;
}

/* Communicate data to JAva application */
void PassSimDataToJava(jstring mes, JNIEnv *env)
{
  (*env)->CallVoidMethod(env, gobj, gJavaCallbackMethodID, mes);
}

/* Read data from C application */
JNIEXPORT void JNICALL Java_Animation_NativeReadTimeUpdate
(JNIEnv *env, jobject obj)
{
  char buf[10000];
  GetJavaHandles(env, obj);
  GetMessage(buf);
  PassSimDataToJava((*env)->NewStringUTF(env, buf) ,env);
}















