/*Copyright (c) 2003 The Regents of the University of California.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. All advertising materials mentioning features or use of this software
   must display the following acknowledgement:
     This product includes software developed by the University of
     California, Los Angeles and its contributors.
4. Neither the name of the University nor the names of its contributors
   may be used to endorse or promote products derived from this software
   without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE. */

import java.io.*;
import java.awt.*;

class Animation {
    int MAXHOSTS;
    final static int NO = 0;
    final static int FROM = 1;
    final static int TO   = 2;
    final static int BOTH = 3;

    static DDoSPanel panel; /* there's a one to one association with the panel
			     * and we need to know which one is ours
			     */
    static Graph graph;	     /* same here */

    public Animation(Graph g, DDoSPanel p, int serverPort, String addresses) 
    {
	MAXHOSTS = 200;
	panel = p;
	graph = g;
	System.out.println("Listening on port "+serverPort);
	NativeInitialize(serverPort,addresses);
    }

    public native int NativeInitialize(int serverPort, String addresses);
    public native void NativeReadTimeUpdate();
  
    static {
	System.loadLibrary("JavaNativeSocket");
    }
  
    public void ReadTimeUpdate() {
	NativeReadTimeUpdate();
    }

    public void UpdateColors(String i)
    {
	if (!i.equals(""))
	    {
		int pos = 0;
		String rest = i;
		while(pos >= 0 && pos < rest.length())
		    {
			pos = rest.indexOf(" ");
			String temp = rest.substring(0,pos);
			rest = rest.substring(pos+1);
			int kind = (new Integer(temp)).intValue();
			if (kind == 1)
			    {
				/* IP Flow message */
				pos = rest.indexOf(" ");
				temp = rest.substring(0,pos);
				int time =  (new Integer(temp)).intValue();
				rest = rest.substring(pos+1);
				pos = rest.indexOf(" ");
				temp = rest.substring(0,pos);
				String source = temp; 
				rest = rest.substring(pos+1);
				pos = rest.indexOf(" ");
				temp = rest.substring(0,pos);
				String dest = temp;
				rest = rest.substring(pos+1);
				pos = rest.indexOf(" ");
				temp = rest.substring(0,pos);
				int sent = (new Integer(temp)).intValue();
				rest = rest.substring(pos+1);
				pos = rest.indexOf(" ");
				temp = rest.substring(0,pos);
				int rec = (new Integer(temp)).intValue();
				rest = rest.substring(pos+1);
				pos = rest.indexOf(" ");
				temp = rest.substring(0,pos);
				int sentp =  (new Integer(temp)).intValue();
				rest = rest.substring(pos+1);
				pos = rest.indexOf(" ");
				temp = rest.substring(0,pos);
				int recp =  (new Integer(temp)).intValue();
				rest = rest.substring(pos+1);
				pos = rest.indexOf(" ");
				temp = rest.substring(0,pos);
				int attp = (new Integer(temp)).intValue();
				rest = rest.substring(pos+1);
				pos = rest.indexOf(",");
				temp = rest.substring(0,pos);
				int att = (new Integer(temp)).intValue();
				panel.handleFlowMessage(time, source, dest, sentp, recp, 
							sent, rec, attp, att);
				System.out.println(time+": Flow message from "+source+" to "+dest+" sent packets "+sentp+" received "+recp+" attack packets "+attp+" attack bytes "+att);
				rest = rest.substring(pos+1);
			    }
			else if (kind == 2)
			    {
				/* Rate message */
				pos = rest.indexOf(" ");
				temp = rest.substring(0,pos);
				int time =  (new Integer(temp)).intValue();
				rest = rest.substring(pos+1);
				pos = rest.indexOf(" ");
				temp = rest.substring(0,pos);
				double ratelim = (new Double(temp)).doubleValue();
				rest = rest.substring(pos+1);
				pos = rest.indexOf(" ");
				temp = rest.substring(0,pos);
				int sentp = (new Integer(temp)).intValue();
				rest = rest.substring(pos+1);
				pos = rest.indexOf(" ");
				temp = rest.substring(0,pos);
				int recp = (new Integer(temp)).intValue();
				rest = rest.substring(pos+1);
				pos = rest.indexOf(" ");
				temp = rest.substring(0,pos);
				int sent = (new Integer(temp)).intValue();
				rest = rest.substring(pos+1);
				pos = rest.indexOf(" ");
				temp = rest.substring(0,pos);
				int rec = (new Integer(temp)).intValue();
				rest = rest.substring(pos+1);
				pos = rest.indexOf(" ");
				temp = rest.substring(0,pos);
				int attp = (new Integer(temp)).intValue();
				rest = rest.substring(pos+1);
				pos = rest.indexOf(" ");
				temp = rest.substring(0,pos);
				int att = (new Integer(temp)).intValue();
				rest = rest.substring(pos+1);
				pos = rest.indexOf(" ");
				temp = rest.substring(0,pos);
				int gooddrop = (new Integer(temp)).intValue();
				rest = rest.substring(pos+1);
				pos = rest.indexOf(" ");
				temp = rest.substring(0,pos);
				int baddrop = (new Integer(temp)).intValue();
				rest = rest.substring(pos+1);
				pos = rest.indexOf(" ");
				temp = rest.substring(0,pos);
				String source = temp;
				rest = rest.substring(pos+1);
				pos = rest.indexOf(",");
				temp = rest.substring(0,pos);
				String node = temp;
				rest = rest.substring(pos+1);
				panel.handleRateMessage(time, source, node, 
							ratelim,sentp, recp, 
							sent, rec, attp, att, gooddrop, baddrop);
				System.out.println(time+": RATE message from "+source+" to "+node+" sent packets "+sentp+" received "+recp+
						   " sent bytes "+sent+" rec bytes "+rec+
						   " ratelim "+ratelim+" attack packets "+attp+" attack bytes "+att+" gooddrop "+gooddrop+" baddrop "+baddrop);
			    }
			else if (kind == 3)
			    {
				/* Interface flow message */
				pos = rest.indexOf(" ");
				temp = rest.substring(0,pos);
				int time =  (new Integer(temp)).intValue();
				rest = rest.substring(pos+1);
				pos = rest.indexOf(" ");
				temp = rest.substring(0,pos);
				String source = temp; 
				rest = rest.substring(pos+1);
				pos = rest.indexOf(" ");
				temp = rest.substring(0,pos);
				int Ninterface = (new Integer(temp)).intValue();
				rest = rest.substring(pos+1);
				pos = rest.indexOf(" ");
				temp = rest.substring(0,pos);
				int sent =  (new Integer(temp)).intValue();
				rest = rest.substring(pos+1);
				pos = rest.indexOf(",");
				temp = rest.substring(0,pos);
				int rec =  (new Integer(temp)).intValue();
				panel.handleFlowMessage(time, source, Ninterface, sent, rec);
				rest = rest.substring(pos+1);
				System.out.println(time+": Interface flow message from "+source+" sent packets "+sent+" received "+rec);
			    }
			else if (kind == 4)
			    {
				System.out.println("RECEIVED MULTICAST");
				pos = rest.indexOf(",");
				temp = rest.substring(0,pos);
				String source = temp; 
				panel.receivedMulticast(source);
				rest = rest.substring(pos+1);
			    }
			else if (kind == 5)
			    {
				System.out.println("SENT MULTICAST");
				pos = rest.indexOf(",");
				temp = rest.substring(0,pos);
				String source = temp; 
				panel.sentMulticast(source);
				rest = rest.substring(pos+1);
			    }
		    }
	    }
    }
}










