/*Copyright (c) 2003 The Regents of the University of California.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. All advertising materials mentioning features or use of this software
   must display the following acknowledgement:
     This product includes software developed by the University of
     California, Los Angeles and its contributors.
4. Neither the name of the University nor the names of its contributors
   may be used to endorse or promote products derived from this software
   without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE. */

import java.util.*;
import javax.swing.*;
import javax.swing.event.*;
import java.awt.event.*;
import java.awt.*;
import java.awt.geom.*;
import java.io.*;

class Node {

    /* Center coordinates */
    double x;
    double y;
    
    /* Width and height */
    double width, height;
    
    /* Is this node rate-limited */
    boolean limited;

    /* This is index in node array */
    int externalID; 

    /* IP address in 32 bit notation */
    long address;
    String lbl;
    SummaryPanel spanel;

    /* Victim, router, normal...and a number */
    String type;
    int ID;
    boolean multicast;
    boolean sent_multicast; 
    int multicast_time;

    /* Is TrackPanel open */
    boolean isTracking;
    TrackPanel panel;
    TrackingFrame tf;
    Node Neighbors[];
    int num_neighbors;
    boolean marked;   
    Vector gooddrops;
    Vector baddrops;
    int total_gooddrop;
    int total_baddrop;
    int total_goodpass;
    int total_badpass;
    String label;

    /* Constructor */
    Node()
    {
	isTracking = false;
	panel = null;
	num_neighbors = 0;
	Neighbors = new Node[10];
	limited = false;
	gooddrops = new Vector(10000);
	baddrops = new Vector(10000);
	multicast = false;
	sent_multicast = false;
	total_gooddrop = total_baddrop 
	    = total_goodpass = total_badpass = 0;
    }

    /* Add drop information */
    void insertDrop(int time, int gooddrop, int baddrop)
    {
	if (time>gooddrops.size())
	    {
		gooddrops.setSize(time+1);
		baddrops.setSize(time+1);
	    }
	if (gooddrops.size()<=(time) 
	    || gooddrops.get(time) == null)
	    {
		if (gooddrop > 0)
		    {
			gooddrops.add(time,new Integer(gooddrop));
			total_gooddrop += gooddrop;
		    }
		if (baddrop > 0)
		    {
			baddrops.add(time,new Integer(baddrop));
			total_baddrop += baddrop;
		    }
	    }
    }

    /* Does node have the given interface */
    public boolean hasNeighbor(int id)
    {
	for(int i=0;i<num_neighbors;i++)
	    if (Neighbors[i].externalID == id)
		return true;
	return false; 
    }
    
    /* Add new interface to the node */
    public void addNeighbor(Node n)
    {
	Neighbors[num_neighbors++] = n;
    }
    
    /* Is this node a victim or not */
    public boolean isVictim()
    {
	if (type.equals("victim"))
	    return true;
	else
	    return false;
    }

    public void receivedMulticast(Graphics g, int time)
    {
	multicast = true;
	sent_multicast = false;
	multicast_time = time;
    }


    public void sentMulticast(Graphics g, int time)
    {
	multicast = true;
	sent_multicast = true;
	multicast_time = time;
    }

    /* Find interface leading to another node */
    public int findNeighbor(Node n)
    {
	for (int i=0;i<num_neighbors;i++)
	    if (Neighbors[i]==n)
  		{	
		    return i;
		}	
	return -1;
    }
    
    /* Find next hop to a given destination */
    int findNextHop(int dest)
    {
	marked = true;
	for(int i=0;i<num_neighbors;i++)
	    {
		if(Neighbors[i].externalID == dest)
		    return Neighbors[i].externalID;
		else if(Neighbors[i].marked==false)
		    {
			if(Neighbors[i].findNextHop(dest) != -1)
			    return Neighbors[i].externalID;
		    }
	    }
	return -1;
    }
    
    /* Draw the node on the panel */
    public void paintNode(Graphics g, Node pick, FontMetrics fm, double xscale,
			  double yscale, double xoffset, double yoffset, 
			  boolean visible, int curtime)
    {
	
	int x = (int) ((double) this.x * xscale);
	int y = (int) ((double) this.y * yscale);
	
	System.out.println("Node "+type+" "+ID+" multicast "+multicast+" sent "+sent_multicast+" time "+multicast_time+" now "+curtime);
	if (type.startsWith("router"))
		g.setColor(new Color(135,206,250));
	else if (type.startsWith("victim"))
	    g.setColor(new Color(250,250,0));
	else if (type.startsWith("ng"))
	    g.setColor(new Color(220,220,220));
	else if (type.startsWith("client"))
	    g.setColor(new Color(152,251,162));
	else if (type.startsWith("attacker"))
	    g.setColor(new Color(255,99,71));	
	else if (type.startsWith("dward"))
	    g.setColor(new Color(55,200,71));	
	else if (type.startsWith("watchdog"))
	    g.setColor(new Color(255,199,171));	
	
	String str_id = pick.lbl; 
	
	int w = fm.stringWidth(type)+ 40;
	int h = 2*(fm.getHeight() + 5);
	
	pick.width = w;
	pick.height = h;
	if (!visible)
	    {
    		g.clearRect(x - w/2, y - h / 2, w, h);
		return;
	    }
	if (type.startsWith("router"))
	    {
		g.fillOval(x - w/2, y - h/2, w, h);
	    }
	else if (type.startsWith("dward"))
	    {
		if (multicast)
		    {
			g.clearRect(x - w/2, y - h / 2, w, h);
			if(curtime>=multicast_time && curtime<=multicast_time+1)
			    {
				if (sent_multicast)
				    g.setColor(Color.yellow);
				else
				    g.setColor(Color.green);
				g.fillOval(x - w/2, y - h/2, 10, 10);
			    }
			else
			    {
				multicast = false;
				g.clearRect(x - w/2, y - h/2, 10, 10);
				g.setColor(new Color(135,206,250));
			    }
			g.setColor(new Color(135,206,250));
		    }
		g.fillOval(x - w/2, y - h/2, w, h);
	    }
	else
	    g.fillRect(x - w/2, y - h/2, w, h);
	g.setColor(Color.black);
	if (type.startsWith("router"))
	{
	    g.drawOval(x - w/2, y - h/2, w-1, h-1);
	}
	else if (type.startsWith("dward") || type.startsWith("watchdog"))
	{
	    g.drawOval(x - w/2, y - h/2, w-1, h-1);
	    if (type.startsWith("dward"))
		g.drawOval(x - w/2, y - h/2, 10, 10);
	}
	else
	    g.drawRect(x - w/2, y - h /2, w-1, h-1);
	g.drawString(label+" "+ID, x - (w-20)/2, 
		     (y - (h/2-5)/2) + fm.getAscent());
  }
}


















