/*Copyright (c) 2003 The Regents of the University of California.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. All advertising materials mentioning features or use of this software
   must display the following acknowledgement:
     This product includes software developed by the University of
     California, Los Angeles and its contributors.
4. Neither the name of the University nor the names of its contributors
   may be used to endorse or promote products derived from this software
   without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE. */

/* File that contains some constants for statistics gathering 
   on aggregate flows to IP address */

#ifndef __FLOW_STATS_H
#define __FLOW_STATS_H

#include "constants.h"
#include "classifier.h"
#include "dward.config"
#include "hash.h" 

/* Flow classification */
enum flow_class {NORMAL, SUSPICIOUS, ATTACK};

struct flow_stats
{
  /* Number of packets received from this address */
  int received_from;		
  /* Number of packets sent to this address */
  int sent_to;			
  /* Number of bytes received from this address */
  int received_bytes;		
  /* Number of bytes sent to this address */
  int sent_bytes;		
  /* Number of attack packets sent to this address, for GUI purpose */
  int attack_packets;			
  /* Number of attack bytes sent to this address, for GUI purpose */
  int attack_bytes;			
  /* Number of forwarded packets */
  int fw_packets;
  /* Number of forwarded bytes */
  int fw_bytes;
  /* Interarrival time between successive packets 
     sent to this address */
  double send_interval;		
  /* Interarrival time between successive packets 
     received from this address */ 
  double receive_interval;	
  /* Last send time within interval to this address */   
  double lst;			
  /* Last receive time within interval from this address */
  double lrt;		
  /* Smoothed sent.rec packet ratio */
  double tcprto;
  double icmprto;
  /* Bytes that are part of good connections */
  int good_bytes;
  /* Time when this entry was last reset */
  long timestamp;
  /* Current flow classification */
  enum flow_class classification;
  /* Old flow classification for statistics gathering */
  enum flow_class old_classification;
  /* Does this flow have the rate limit,
     for statistics gathering */
  int limited;
  /* Compliant period */
  int compliant;
  /* Per protocol statistcs */
  int TCP_received_from;
  int TCP_sent_to;
  int TCP_received_bytes;
  int TCP_sent_bytes;
  int UDP_received_from;
  int UDP_sent_to;
  int UDP_received_bytes;
  int UDP_sent_bytes;
  int ICMP_received_from;
  int ICMP_sent_to;
  int ICMP_received_bytes;
  int ICMP_sent_bytes;
  int OTHER_received_from;
  int OTHER_sent_to;
  int OTHER_received_bytes;
  int OTHER_sent_bytes;
  /* Smoothed ratio of various protocols */
  double tcp;
  double udp;
  double icmp;
  double other;
  /* Number of connections per protocol */
  int TCP_conns;
  int UDP_conns;
  int ICMP_conns;
  int OTHER_conns;
  /* INC statistics */
  float min_TCP;
  float min_UDP;
  float min_ICMP;
  float min_OTHER;
  float track_TCP;
  float track_UDP;
  float track_ICMP;
  float track_OTHER;
  float stats[WINDOW];
  int avg_time;
  int full;
  int pos;
  /* DefCOM fields */
  int DefCOM_attack;
  int DefCOM_rate;
  /* COSSACK detected attack flag */
  int COSSACK_attack;
  int COSSACK_rate;
  int COSSACK_type;
  int COSSACK_new;
  int COSSACK_timestamp;
 /* For red-teaming */
  int ID;
  /* For debugging */
  int observe;
};

/* Function prototypes, refer to implementation for descriptions */
int flow_clean(struct flow_stats*);
int flow_compare(unsigned int*, unsigned int*);
unsigned long flow_hash1(unsigned int*, unsigned long);
unsigned long flow_hash2(unsigned int*, unsigned long);
void clean_flow_hash();
long find_flow(char *);
long find_flow_num(unsigned int);
long add_flow(char*);
long add_flow_num(unsigned int);
void update_flow(char*, enum direction, int, int, int, int,enum pack_type);
void update_flow_num(unsigned int, enum direction, int,int, 
		     int, int, enum pack_type);
void initialize_flows();
void flowhash_cleanup(int);
int find_flow_class(unsigned int);
void reset_flow(long);
void update_connections(unsigned int, int);
void remove_connections(unsigned int, int);
int set_cossack_flag_and_rate(unsigned int, int, int, int);
int set_alert(unsigned int);
int set_rate_limit(unsigned int, int);
void monitor_flow(unsigned int);

#endif



