/* Copyright (c) 2003 The Regents of the University of California.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. All advertising materials mentioning features or use of this software
   must display the following acknowledgement:
     This product includes software developed by the University of
     California, Los Angeles and its contributors.
4. Neither the name of the University nor the names of its contributors
   may be used to endorse or promote products derived from this software
   without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.*/

/* System includes */
#include <linux/time.h>
#include <linux/config.h>  
#include <linux/slab.h> 
#include <linux/mm.h> 
#include <linux/module.h>
#include <linux/kernel.h>
#include <asm/types.h> 
#include <linux/stddef.h> 
#include <linux/types.h> 
#include <linux/netfilter.h> 
#include <linux/netfilter_ipv4.h> 
#include <linux/skbuff.h> 
#include <linux/in.h> 
#include <linux/ip.h> 
#include <linux/tcp.h> 
#include <linux/udp.h> 
#include <linux/icmp.h> 
#include <linux/netdevice.h> 
#include <linux/if_ether.h> 
#include <sys/syscall.h>  
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/types.h> 
#include <linux/netfilter.h> 
#include <linux/netfilter_ipv4.h> 
#include <linux/skbuff.h>
#include <linux/unistd.h>
#include <asm/semaphore.h>
#include <endian.h>
#include <unistd.h>
#include <linux/devfs_fs_kernel.h>

#include "hash.h"
#include "fhash.h"
#include "chash.h"
#include "constants.h"
#include "dward.h"
#include "machines.h"
#include "dward.config"

/* Command line initialization parameter */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,2,0)
MODULE_PARM(drop, "i");
MODULE_PARM(mark, "i");
MODULE_PARM(defcom,"i");
MODULE_PARM(hm,"i");
MODULE_PARM(LOCAL_ADDRESS, "s");
MODULE_PARM(LOCAL_MASK, "i");
#endif

extern void* sys_call_table[];
extern void create_flowhash();
extern void create_connhash();
extern struct hash_table* ConnHash; 
struct timer_list *my_timer;
extern struct file_operations dward_fops;


/* How many dropped packets to log on attack signal */
#define MAX_LOG 100
int dropped = MAX_LOG;

/* Local address and mask */
char *LOCAL_ADDRESS = 0;
int LOCAL_MASK = 0;

/* Interface with DefCOM */
int defcom = FALSE;

/* Host models */
int hm = FALSE;
struct machine my_hosts[MAX_MACHINES];
int my_num_machines = 0;

/* Should this module drop any packets. 
   This flag can be set from command line */
int drop = TRUE;

/* The attack packet's value in TOS field, for statistics 
   gathering purposes. This flag can be set from command line */
int mark = 255;

/* Some statistics */
int totalmem = 0;
int attack_passed[STAT_SIZE], good_passed[STAT_SIZE], attack_dropped[STAT_SIZE], good_dropped[STAT_SIZE];
int forwarded[STAT_SIZE],goodB[STAT_SIZE];
int beginstats = 0,curstats = 0, times = 0;
int fw = 0;

/* For exclusive access to data */
struct semaphore s;
struct timeval tv, now,pv;
__u32 temp;

/* Make dotted notation from 32-bit address */
extern void ntoa( __u32, char*);

/* Make 32-bit address from dotted notation */
__u32 addr(char* c)
{
  int i, temp = 0;
  __u32 address = 0;
  for(i=0;i<17;i++)
    {
      /* End of string */
      if (c[i] == 0)
	break;
      if (c[i]!='.')
	temp = temp*10 + (c[i]-'0');
      else
	{
	  if (temp < 256)
	    {
	      address = address * 256 + temp;
	      temp = 0;
	    }
	  else
	    return 0; 
	}
    }
  address = address * 256 + temp;
  return address;
}

/* Getflow system call, returns statistics on given flow */
int getflow(unsigned int address, int type)
{
  int k=get_flow(address, type);
  return k;
}

/* Getconn system call, returns statistics on given connection */
int getconn(unsigned int address1, 
		unsigned int address2,
		unsigned short port1, unsigned short port2, int type)
{
  return get_conn(address1,address2,port1,port2, type);
}

/* Resetflow system call, resets flow statistics */
int resetflow(unsigned int address)
{
  reset_flow(address);
  return 0;
}

/* Resetconn system call, resets connection statistics */
int resetconn(unsigned int address1, 
		  unsigned int address2,
		  unsigned short port1, unsigned short port2)
{
  reset_conn(address1,address2,port1,port2);
  return 0;
}

/* Addflow system call, inserts new flow */
int addflow(unsigned int address, int rl)
{
  insert_rate_limit(address, rl);
  dropped = MAX_LOG;
  return 0;
}

/* Delflow system call, deletes flow */
int delflow(unsigned int address)
{
  delete_rate_limit(address);
  return 0;
}

/* Addconn system call, inserts new connection */
int addconn(unsigned int address1, 
	    unsigned int address2,
	    unsigned short port1, unsigned short port2, 
	    int protocol)
{
  int i=insert_connection(address1,address2,port1,port2,protocol);
  /* Validate connection
     if we are using host models */
  if (i!=NOT_FOUND && i!=-2 && hm && protocol == IPPROTO_TCP)
    {
      int i,k;
      for(i=0;i<my_num_machines;i++)
	{
	  if(my_hosts[i].address == address1)
	    if ((k=find_port(&my_hosts[i],port1))!= NOT_FOUND)
		    {
		      /* Validate already inserted record for goodness */
		      validate_data(&my_hosts[i],k);
		      /* Try to classify the machine */
		      classify_machine(&my_hosts[i]);
		    }
	}
    }
  return 0;
}

/* Delconn system call, deletes connection */
int delconn(unsigned int address1, 
	    unsigned int address2,
	    unsigned short port1, unsigned short port2)
{
  delete_connection(address1,address2,port1,port2);
  /* Remove port info from the machine
     if we are using host models */
  if (hm)
    {
      int i;
      for(i=0;i<my_num_machines;i++)
	{
	  if(my_hosts[i].address == address1)
	    delete_port(&my_hosts[i],port1);
	}
    }
  return 0;
}

/* Getstats system call, gets packet statistics from kernel */
int getstats(int type)
{
  int val;
  if (beginstats != curstats)
    {
      switch(type)
	{
	case GOOD_DROP:
	  val = good_dropped[beginstats];
	  good_dropped[beginstats] = 0;
	  break;
	case GOOD_PASS:
	  val = good_passed[beginstats];
	  good_passed[beginstats] = 0;
	  break;
	case BAD_DROP:
	  val = attack_dropped[beginstats];
	  attack_dropped[beginstats] = 0;
	  break;
	case BAD_PASS:
	  val = attack_passed[beginstats];
	  attack_passed[beginstats] = 0;
	  break;
	case FORWARDED:
	  val = forwarded[beginstats];
	  forwarded[beginstats] = 0;
	  break;
	case GOODPUT:
	  val = goodB[beginstats];
	  goodB[beginstats] = 0;
	  break;
	default:
	  return -1;
	}
      if (times == 5)
	{
	  beginstats = (beginstats + 1) % STAT_SIZE;
	  times = 0;
	}
      else
	times ++;
      return val;
    }
  else
    return NOT_FOUND;
}

/* Re-initialize timer */
void clean_counter(unsigned long n)
{
  kfree(my_timer);
  reset_flows();
  reset_conns();
  my_timer = kmalloc(sizeof(struct timer_list), GFP_KERNEL);
  init_timer(my_timer);
  my_timer->data = 0;
  my_timer->expires = jiffies+1*HZ;
  my_timer->function = clean_counter;
  down(&s);
  /* Shift statistics */
  curstats = (curstats + 1) % STAT_SIZE;
  up(&s);
  add_timer(my_timer);
}

/* Filtering function, called whenever a packet is intercepted */
static unsigned int nf_packet_passing(unsigned int hook, struct sk_buff **pskb,
				      const struct net_device *indev, 
				      const struct net_device *outdev, 
				      int (*okfn)(struct sk_buff *)) 
{ 
  long sindex, dindex;
  int sport = -1, dport = -1,bytes;
  struct iphdr *ip;
  struct tcphdr *tcp;
  struct udphdr* udp;
  struct icmphdr *icmp;
  __u32 source,dest,seqnum = 0, acknum = 0;
  int k, rate_limit, sent_bytes, good_bytes, gbsec, i, j, syns_sent;
  char source_string[18], dest_string[18];

  do_gettimeofday(&pv);
  down(&s);
  ip = (*pskb)->nh.iph;
  tcp = (struct tcphdr*)((__u32*)ip+ip->ihl);
  udp =(struct udphdr*)((__u32*)ip+ip->ihl);
  icmp = (struct icmphdr*)((__u32*)ip+ip->ihl);

  source = ip->saddr;
  dest = ip->daddr;
  
  sindex = find_flow(source);
  dindex = find_flow(dest);
  bytes = ntohs(ip->tot_len);

  forwarded[curstats] += bytes;

  /* Take out ports */
  if (ip->protocol == IPPROTO_IP || ip->protocol == IPPROTO_TCP)
    { 
      sport = ntohs(tcp->source);
      dport = ntohs(tcp->dest);
      seqnum = (__u32)ntohl(tcp->seq);
      acknum = (__u32)ntohl(tcp->ack);
      /* If this is good connection and we are building
	 host models, update the model */
      if(hm)
	{
	  for(i=0;i<my_num_machines;i++)
	    {
	      /* If I already have machine info */
	      if(((source & temp) == (htonl(addr(LOCAL_ADDRESS)) & temp)) 
		 && (my_hosts[i].address == ip->saddr))
		{
		  /* If this is new connection */
		  if (((k=find_port(&my_hosts[i],ntohs(tcp->source))) == NOT_FOUND) && (tcp->syn))
		    {
		      /* create new record */
		      insert_data(&my_hosts[i],ntohs(tcp->source),
				  ntohl(tcp->seq));
		    }
		  break;
		}
	    }
	  /* Else add new host if it is from our network */
	  if ((((source & temp) == (htonl(addr(LOCAL_ADDRESS)) & temp)))
	      && (i == my_num_machines) && (my_num_machines < MAX_MACHINES))
	    {
	      my_hosts[i].address = ip->saddr;
	      init_machine(&my_hosts[i]);
	      /* create new record, only for SYN packets */
	      if (tcp->syn)
		insert_data(&my_hosts[i],ntohs(tcp->source),ntohl(tcp->seq));
	      my_num_machines++;
	    }
	}  
      
    }
  else if (ip->protocol == IPPROTO_UDP)
    {
      sport = htons(udp->source);
      dport = htons(udp->dest);
      seqnum = 0;
      acknum = 0;
    }
  else if (ip->protocol == IPPROTO_ICMP)
    {
      sport = 0;
      dport = 0;
      seqnum = 0;
      acknum = 0;
    }
  else printk("unknown proto %d\n",ip->protocol);

  /* If this is incoming packet pass it */
    if ((dest & temp) 
    == (htonl(addr(LOCAL_ADDRESS)) & temp)) 
      {
	up(&s);
	do_gettimeofday(&tv);
	return NF_ACCEPT;
      }
    else
      {
	/* Outgoing packet */
	/* If it is randomly spoofed, drop it */
	if ((source & temp)
            != (htonl(addr(LOCAL_ADDRESS)) & temp))
      	{
           if (ip->tos == mark)
              attack_dropped[curstats] += bytes;
            else
                good_dropped[curstats] += bytes;

	   ntoa(source, source_string);
	   ntoa(dest, dest_string);
	  printk("Dropping spoofed packet from %s:%u to %s:%u protocol %d\n",source_string,sport,dest_string,dport,ip->protocol); 

       up(&s);
        do_gettimeofday(&tv);
        return NF_DROP;
       }

	/* If it is not limited flow  */
	if (dindex == NOT_FOUND)
	  {
	    /* Update outgoing packet statistics */
	    if (ip->tos == mark)
	      attack_passed[curstats] += bytes;
	    else
	      {
		good_passed[curstats] += bytes;
		goodB[curstats] += get_good(source,dest,sport,dport);
	      }
	    up(&s);
	    do_gettimeofday(&tv);
	    return NF_ACCEPT;
	  }
	else
	  {
	    /* Rate limited flow */
	    rate_limit = get_flow(dest,RL);
	    syns_sent = get_flow(dest,SYNS);
	    sent_bytes = get_flow(dest,SENT_BYTES_SEC);
	    good_bytes = get_flow(dest,GOOD_BYTES);
	    gbsec = get_flow(dest,GOOD_BYTES_SEC);
	    
	    /* Good connection, always pass packets */
	    if ((k = find_connection(source,dest,sport,dport))!=NOT_FOUND)
	      {
		update_flow(dest,bytes,OUT,ip->tos!=mark,FALSE);
		update_flow_good(dest,bytes);
		if (ip->tos == mark)
		  attack_passed[curstats] += bytes;
		else
		  {
		    good_passed[curstats] += bytes;
		    goodB[curstats]+=get_good(source,dest,sport,dport);
		  }
		/* Defcom extension */
  		if (defcom)
		{
			ip->id=1;
			ip->check=0;
		}
		up(&s);
		do_gettimeofday(&tv);
		return NF_ACCEPT;
	      }
	    else 
	      {
		/* Transient connection, let it through only if good 
		   connections have 
		   sufficient remaining bandwidth */
		struct timeval tv;
		__u32 prediction;
		int gb, good_packet = FALSE;
		do_gettimeofday(&tv);
		gb = 1.5*good_bytes;

		/* First apply host models if we are using them 
		   but only for TCP packets */
		if(hm && (ip->protocol == IPPROTO_TCP))
		  {
		    int i;
		    for(i=0;i<my_num_machines;i++)
		      {
			/* If have machine info and it is classified */
			if(my_hosts[i].address == ip->saddr)
			{
			  if(my_hosts[i].os_type != NOS)
			    {
			      prediction = 
				predict_seqnum(&my_hosts[i],my_hosts[i].prev,
					       tv.tv_sec,my_hosts[i].os_type);
			      if(verify_prediction(prediction,ntohl(tcp->seq),
						   my_hosts[i].os_type))
				{
				  if(syns_sent+bytes < SYN_RATE)
				    {
				      /* Pass packet */
				      good_packet = TRUE;
				    }
			      }
			    }
			  break;
			}
		      }
		  }
		if (((rate_limit > 0) 
		     && (((sent_bytes+bytes) < (rate_limit)) 
			 && ((sent_bytes-gbsec+bytes) < (rate_limit - gb)))) 
		    || ((rate_limit <= 0) || (drop == FALSE) || good_packet))
		  {
		    update_flow(dest,bytes,OUT,ip->tos!=mark, good_packet);
		    if (ip->tos == mark)
		      attack_passed[curstats] += bytes;
		    else
		      {
			good_passed[curstats] += bytes; 
			goodB[curstats] += get_good(source,dest,sport,dport);
		      }
		    /* Defcom extension */
		    if (defcom)
                   {
		    	ip->id=2;
		    	ip->check=0;
		    	//ip->check=ip_fast_csum((unsigned char *)ip, ip->ihl);
		    }
		    up(&s);
		    do_gettimeofday(&tv);
		    return NF_ACCEPT;
		  }
		else
		  {
		    update_flow_drop(dest, bytes, FALSE);
		    if (ip->tos == mark)
		      attack_dropped[curstats] += bytes;
		    else
		      {
			if (dropped)
			  {
			    ntoa(source, source_string);
			    ntoa(dest, dest_string);
			    printk("Dropping packet from %s:%u to %s:%u protocol %d\n",source_string,sport,dest_string,dport,ip->protocol); 
			    dropped --;
			  }
			good_dropped[curstats] += bytes;
		      }
		    up(&s);
		    do_gettimeofday(&tv);
		    return NF_DROP;
		  }
	      }
	  }
      } 
}

static struct nf_hook_ops forward_chain = { { NULL, NULL }, nf_packet_passing, PF_INET, NF_IP_FORWARD, 0 }; 

/* Initialization function, called when a module is loaded */
int init_module()
{
  int retval,i;
  printk("Instaled rl module\n");
  if (!LOCAL_ADDRESS)
  {	
	printk("LOCAL ADDRESS not set\n");
	return -EIO;
  }
  temp = ~(((__u32)(~0)) << (__u32)(LOCAL_MASK));
  init_MUTEX(&s);
  my_timer = kmalloc(sizeof(struct timer_list), GFP_KERNEL);
  init_timer(my_timer);
  my_timer->data = 0;
  my_timer->expires = jiffies+1*HZ;
  my_timer->function = clean_counter;
  add_timer(my_timer);
  do_gettimeofday(&tv);
  create_flowhash();
  create_connhash();
  /* Open D-WARD device */
  if (devfs_register_chrdev (DWARD_MAJOR, "dward", &dward_fops)) {
    printk ("dward: unable to get major %d\n", DWARD_MAJOR);
    return -EIO;
  }
  nf_register_hook(&forward_chain);  
  /* Initialize statistics */
  for(i=0;i<STAT_SIZE;i++)
    {
      attack_passed[i] = 0;
      good_passed[i] = 0; 
      attack_dropped[i] = 0; 
      good_dropped[i] = 0;
      forwarded[i] = 0;
      goodB[i] = 0;
    }
  return 0;
}

/* Restore function, called when a module is unloaded */
void cleanup_module()
{
  del_timer(my_timer);
  devfs_unregister_chrdev(DWARD_MAJOR, "dward");
  nf_unregister_hook(&forward_chain); 
}










