/* Copyright (c) 2003 The Regents of the University of California.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:

1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. All advertising materials mentioning features or use of this software
   must display the following acknowledgement:
     This product includes software developed by the University of
     California, Los Angeles and its contributors.
4. Neither the name of the University nor the names of its contributors
   may be used to endorse or promote products derived from this software
   without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
 */

/* Hash table implementation */
#include "constants.h"
#include "hash.h"
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>
#include <limits.h>

int hashmem = 0;

/* This is also defined in khash.c, so if changed should be changed
   in both files */
static unsigned long prime_list[]={
  2,3,5,11,23,53,73,101,127,151,173,211,223,251,271,307,331,353,373,401,
  421,449,479,503,523,557,571,601,631,653,673,701,727,751,773,809,823,
  853,877,907,929,953,977,1009,1097,1201,1301,1409,1511,1601,1709,1801,
  1901,2003,2111,2203,2309,2411,2503,2609,2707,2801,2903,3001,3109,3203, 
  3301,3407,3511,3607,3701,3803,3907,4001,4111,4201,4327,4409,4507,4603,
  4703,4801,4903,5003,5101,5209,5303,5407,5501,5591,5701,5801,5903,6007,
  6101,6203,6301,6397,6521,6607,6701,6803,6907,7001,7103,7207,7307,7411,
  7507,7603,7703,7817,7901,8009,8101,8209,8311,8419,8501,8599,8699,8803,
  8923,9001,9103,9203,9311,9403,9511,9601,9719,9803,9901,10007,11003,12049,
  13033,14009,15013,16001,17011,18013,19001,20011,21089,22003,23003,24001,
  25013,26003,27011,28001,29009,30011,31013,32003,33029,34019,35023,36011,
  36943,37963,38993,40009,41051,42013,43003,44087,45083,46147,47161,48179,
  49169,50119,51001,52009,53069,54059,55001,56003,57037,58013,59023,60029,
  61091,62137,63277,64223,65089,66029,67103,68059,69193,70249,71059,72031,
  73039,74161,75011,76031,77153,78139,79043,80077,81049,82073,83203,84589,
  85447,86477,87281,88379,89101,90071,91159,92083,93053,94111,95153,96181,
  97301,98041,99109,100003};
static int prime_list_length = 222;		 

/* Function that tests if number is prime */
int test_prime(long num)
{
  long i;
  for(i=2;i<num;i++)
    if (num % i == 0)
      return FALSE;
  return TRUE;
}

/* Creating hash table 
   input: object to store the hash table in,
          desired size, minimum and maximum size,
	  and load factors for rehashing, key size,
          function that decides whether to clean the entry,
	  function that compares entries,
	  first and second hash function */
void create_hash(struct hash_table** HT, unsigned long size, 
		 unsigned long min_size, unsigned long max_size, 
		 double rlfh, double rlfl, int key_size, int data_size,
		 int (*cf)(void*), int (*cmp)(void*,void*),
		 unsigned long (*hash1)(void*,unsigned long), 
		 unsigned long (*hash2)(void*,unsigned long))
{
  int i;
  
  struct hash_table* HashTable;

  /* Allocate space */
  HashTable = malloc(sizeof(struct hash_table));
  hashmem += (int)sizeof(struct hash_table);

  /* Test if desired sizes are prime and if not
     find prime numbers closest to the values that are given */

  if (!test_prime(size))
    {
      for(i=0;i<prime_list_length && size>prime_list[i];i++);
      if (i != prime_list_length)
	size = prime_list[i];
    }

  if (!test_prime(min_size))
    {
      for(i=0;i<prime_list_length && min_size>prime_list[i];i++);
      if (i != prime_list_length)
	min_size = prime_list[i];
    }

  if (!test_prime(max_size))
    {
      for(i=0;i<prime_list_length && max_size>prime_list[i];i++);
      if (i != prime_list_length)
	max_size = prime_list[i];
    }

  HashTable->entries = malloc(size*sizeof(struct hash_entry));
  hashmem += size*sizeof(struct hash_entry);
  
  /* Set all keys to EMPTY */
  HashTable->size = size;
  HashTable->filled = 0;
  HashTable->rehash_lf_high = rlfh;
  HashTable->rehash_lf_low = rlfl;
  HashTable->FULL_ALERT = FALSE;
  HashTable->data_size = data_size;
  HashTable->key_size = key_size;
  HashTable->min_size = min_size;
  HashTable->max_size = max_size;
  HashTable->clean = cf;
  HashTable->compare = cmp;
  HashTable->hash1 = hash1;
  HashTable->hash2 = hash2;
  HashTable->stamp = 0;
  HashTable->no_rehash = TRUE;
  for(i=0;i<size;i++)
    {
      HashTable->entries[i].key = malloc(HashTable->key_size);
      HashTable->entries[i].data = malloc(HashTable->data_size);
      HashTable->entries[i].state = EMPTY;
      HashTable->entries[i].visited = 0;
    }
  *HT = HashTable;
}

/* Function that cleans unnecessary entries */
void clean_table(struct hash_table* HashTable)
{
  int i;
  for(i=0;i<HashTable->size;i++)
      if (HashTable->entries[i].state == VALID && 
	  HashTable->clean(HashTable->entries[i].data))
	{
	  HashTable->entries[i].state = DELETED;
	  HashTable->filled--;
	}
}

/* Rehash function
   input: int (1 - enlarge table, 0 - shrink table) 
          hash table to rehash */
void rehash(int enlarge, struct hash_table* HashTable)
{
  int i;

  /* Save old size */
  unsigned long old_size = HashTable->size;
  struct hash_entry* temp_table;

  if (enlarge)
    {
      /* Rehash should enlarge the table */
      if (HashTable->size == HashTable->max_size)
	return;
      /* We should enlarge the table, allocate almost double space */
      if (2*HashTable->size >= HashTable->max_size)
	{
	  /* We need to clean the table since it can not
	     grow any bigger */
	  clean_table(HashTable);
	  HashTable->size = HashTable->max_size;
	}
      else
	{
	  HashTable->size = 2 * HashTable->size;
	  /* Find closest prime number to desired size */
	  for(i=0;i<prime_list_length && HashTable->size>prime_list[i];i++);
	  if (i<prime_list_length)
	    HashTable->size = prime_list[i];
	}
      /* save old contents */
      temp_table = HashTable->entries;
      
      /* Allocate new space */
      HashTable->filled = 0;
      HashTable->entries 
	= malloc(HashTable->size*sizeof(struct hash_entry));
      hashmem += (HashTable->size*sizeof(struct hash_entry));

      /* Reset all entries */
      for(i=0;i<HashTable->size;i++)
	{
	  HashTable->entries[i].key = malloc(HashTable->key_size);
	  HashTable->entries[i].data = malloc(HashTable->data_size);
	  HashTable->entries[i].state = EMPTY;
	  HashTable->entries[i].visited = 0;
	}

      for(i=0;i<HashTable->size;i++)
	{
	  /* Re-insert values from old table */
	  if (i<old_size && temp_table[i].state != EMPTY && 
	      temp_table[i].state != DELETED)
	      {
		int outcome;
		int index = insert(temp_table[i].key, HashTable,&outcome);
		memcpy(HashTable->entries[index].data,
		       temp_table[i].data, HashTable->data_size);
		free(temp_table[i].key);
		free(temp_table[i].data);
		hashmem -= HashTable->key_size;	  

		temp_table[i].key = NULL;
		temp_table[i].data = NULL;
	      }
	}
    }
  else
    {
      /* Rehash should reduce the table */
      if (HashTable->size == HashTable->min_size)
	return;

      /* We should shrink the table, allocate the space 
         that will not lead to immediate enlargement of 
         hash table (i.e. dependent on rehash_lf_high) */
      if (HashTable->filled/HashTable->rehash_lf_high * 1.1 
	  < HashTable->min_size)
	  HashTable->size = HashTable->min_size;
      else
	HashTable->size 
	  = HashTable->filled/HashTable->rehash_lf_high * 1.1;

      /* Find closest prime number to desired size */
      for(i=0;i<prime_list_length && HashTable->size>prime_list[i];i++);
      if (i<prime_list_length)
	HashTable->size = prime_list[i];
      /* save old contents */
      temp_table = HashTable->entries;
      /* Allocate new table */
      HashTable->filled = 0;
      HashTable->entries = 
	malloc(HashTable->size*sizeof(struct hash_entry));
      hashmem += (HashTable->size*sizeof(struct hash_entry));

      /* Reset all entries */
      for(i=0;i<HashTable->size;i++)
	{
	  HashTable->entries[i].key = malloc(HashTable->key_size);
	  HashTable->entries[i].data = malloc(HashTable->data_size);
	  HashTable->entries[i].state = EMPTY;
	  HashTable->entries[i].visited = 0;
	}
      for (i=0;i<old_size;i++)
	{
	  /* Re-insert values from old table */
	  if (temp_table[i].state != EMPTY && 
	      temp_table[i].state != DELETED)
	        {
		  int outcome;
		  int index = insert(temp_table[i].key, HashTable, &outcome);
		  memcpy(HashTable->entries[index].data,
			 temp_table[i].data, HashTable->data_size);
		  free(temp_table[i].key);
		  free(temp_table[i].data);
		  hashmem -= HashTable->key_size;	  

		  temp_table[i].key = NULL;
		  temp_table[i].data = NULL;
		}
	}
    }
  /* Delete old table */
  free(temp_table);
  hashmem -= (old_size*sizeof(struct hash_entry));	  
  temp_table = NULL;
}

/* Insert new key
   input: void* key, hash table to modify
          void* data to insert
   output: long index where key is stored, -1 if
           empty slot can not be found */   
long insert(void* key, struct hash_table* HashTable, int* outcome)
{
  unsigned long index;
  int trial = 0;

  /* If hash is full, give up */
  if (HashTable->filled == HashTable->max_size)
    return NOT_FOUND;
  /* Increment value to put into visited field */
  if (HashTable->stamp < LONG_MAX)
    HashTable->stamp++;
  else
    HashTable->stamp = 1;

  /* Keep looking for appropriate empty slot */
  do{
    index = ((HashTable->hash1(key, HashTable->size) 
	      + trial * HashTable->hash2(key, HashTable->size))
	     % HashTable->size);
    trial++;
    /* FIX: limit trials to avoid colissions */
    if (trial > 3)
      return NOT_FOUND;

    /* check if this is second visit to the same slot,
       if it is this means that either hash is full or
       hash function is not well-designed and we are in the 
       infinite loop. Then return NOT FOUND */
    if (HashTable->entries[index].state == VALID
	&& !HashTable->compare(HashTable->entries[index].key, key)
	&& HashTable->entries[index].visited == HashTable->stamp)
      return NOT_FOUND;
    HashTable->entries[index].visited = HashTable->stamp;
  } while(HashTable->entries[index].state != EMPTY && 
	  HashTable->entries[index].state != DELETED && 
	  !HashTable->compare(HashTable->entries[index].key, key));

  /* Eliminate possibility of double insertion */
  if (HashTable->entries[index].state == VALID && 
      HashTable->compare(HashTable->entries[index].key, key))
    {
      *outcome = EXISTING;
      return index;
    }
  /* Insert key */
  memcpy(HashTable->entries[index].key, key, HashTable->key_size);
  HashTable->entries[index].state = VALID;
  HashTable->filled++;
  *outcome = NEW;

  /* Set FULL_ALERT flag if hash is more than 90% full */
  if (HashTable->size == HashTable->max_size 
      && (double)HashTable->filled/HashTable->size > FULL_ALERT_LF)
        HashTable->FULL_ALERT = TRUE;
  else
    HashTable->FULL_ALERT = FALSE;
  /* Check if we need to rehash */
  if ((double)HashTable->filled/HashTable->size > HashTable->rehash_lf_high &&
      !HashTable->no_rehash)
  {
    rehash(1, HashTable);
    return find(key, HashTable);
  }
  else
    return index;
}

/* Find a key
   input: void* key, hash table to search
   output: long index where key is stored, 
           NOT_FOUND is returned if key is not in the table */   
long find(void* key, struct hash_table* HashTable)
{
  unsigned long index;
  int trial = 0;
 
 /* Increment value to put into visited field */
  if (HashTable->stamp < LONG_MAX)
    HashTable->stamp++;
  else
    HashTable->stamp = 1;

  /* Keep looking for the key */
  do{
      index = (HashTable->hash1(key,HashTable->size) 
	       + trial * HashTable->hash2(key,HashTable->size))
	       % HashTable->size;
      trial++;
 
       if(trial > 4)
	 return NOT_FOUND;

    /* check if this is second visit to the same slot,
       if it is this means that either hash is full or
       hash function is not well-designed and we are in the 
       infinite loop. Then return NOT FOUND */
      if (HashTable->entries[index].visited == HashTable->stamp)
	return NOT_FOUND;
      HashTable->entries[index].visited = HashTable->stamp;
  } while(HashTable->entries[index].state != EMPTY && 
	  !HashTable->compare(HashTable->entries[index].key, key));
	  
  /* Return index where key is found or NOT_FOUND */
  if (HashTable->entries[index].state == VALID && 
      HashTable->compare(HashTable->entries[index].key, key))
      return index;
  else
      return NOT_FOUND;
}

/* Delete a key
   input: void* key, hash table to modify
   output: long index where key was stored, 
           NOT_FOUND is returned if key is not in the table */   
long delete(void* key, struct hash_table* HashTable)
{
  long index;

  index = find(key, HashTable);
  if (index != NOT_FOUND)
  {
    /* Delete the key. Reson why a special value DELETED is used
       here, instead of EMPTY is to enable access to values stored
       after collision value that is being deleted. This space is 
       then available for re-use but does not stop search process */
    HashTable->entries[index].state = DELETED;
    HashTable->filled--;
    
    /* See if we need to rehash */
    if ((double)HashTable->filled/HashTable->size < HashTable->rehash_lf_low 
	&& HashTable->size > HashTable->min_size && !HashTable->no_rehash)
      {
	rehash(0, HashTable);
	return index;
      }
  }
  return index;
}

/* If rehashing was forbidden it is good
   to check whether it should be done
   once it is enabled */
void check_rehash(struct hash_table* HashTable)
{
 if ((double)HashTable->filled/HashTable->size < HashTable->rehash_lf_low 
     && HashTable->size > HashTable->min_size && !HashTable->no_rehash)
          rehash(0, HashTable);
 else 
   if ((double)HashTable->filled/HashTable->size > HashTable->rehash_lf_high 
       && !HashTable->no_rehash)
          rehash(1, HashTable);
}




